bl_info = {
    "name": "Rotate Edge Loop",
    "blender": (3, 0, 0),
    "category": "Object",
    "version": "1.7",
    "author" : "palmpix - pixe.palm@gmail.com",
    "description": "This addon lets you place rotate and slide edge loops",
}

import bpy
import numpy as np
import bmesh
import mathutils as mu
import math

class ModifyGroup(bpy.types.Operator):
    """Modify Vertex Group"""
    bl_idname = "transform.edge_rotate"
    bl_label = "Rotate Edge Loop"
    bl_options = {'REGISTER', 'UNDO'}

    save_main_settings: bpy.props.BoolProperty(
        name="Save Main Settings:",
        description="Keep the same settings for the next edit",
        default = False,
    )   
    save_noise_settings: bpy.props.BoolProperty(
        name="Save Noise Settings:",
        description="Keep the same settings for the next edit",
        default = False,
    )   
    offset_fac: bpy.props.FloatProperty(
        name="Position Offset:",
        description="Position offset of the loop",
        default = 0,
 
        #options={'SKIP_SAVE'}
    )   
    rot_fac: bpy.props.FloatProperty(
        name="Primary Angle:",
        description="Angle of the loop",
        default = 0,

        #options={'SKIP_SAVE'}
    )    
    rot_fac_2: bpy.props.FloatProperty(
        name="General Rotation:",
        description="Offset of the rotation of the loop. It serves well as a quick option in some cases but in others I suggest using the custom rotation",
        default = 0,

        #options={'SKIP_SAVE'}
    )    

    rot_vec: bpy.props.FloatVectorProperty(
        name="Rotation Tweak:",
        description="Tweak rotation of the loop. This will rotate the current loop already rotated by the Generat Rotation. If it is tricky to find a desired rotation I suggest using the General Rotation",
        default = (0,0,0),

        #options={'SKIP_SAVE'}
    )

    test_fac: bpy.props.FloatProperty(
        name="Non-Flat Offset (Experimental):",
        description="Warning: This feature is experimental. This feature will maybe allow you to make Non-Planar loops combined with the Position Offset",
        default = 0,
 
        #options={'SKIP_SAVE'}
    ) 

    noise_offset: bpy.props.FloatProperty(
        name="Noise Influence:",
        description="",
        default = 0,
 
        #options={'SKIP_SAVE'}
    ) 

    noise_scale: bpy.props.FloatProperty(
        name="Noise Scale:",
        description="",
        default = 1,
 
        #options={'SKIP_SAVE'}
    ) 
    noise_vec: bpy.props.FloatVectorProperty(
        name="Noise Position:",
        description="",
        default = (0,0,0),

        #options={'SKIP_SAVE'}
    )
    H: bpy.props.FloatProperty(
        name="Increment Factor:",
        description="",
        default = 0.5,
    )   
    lacunarity: bpy.props.FloatProperty(
        name="Lacunarity:",
        description="",
        default = 0.5,
    ) 
    octaves: bpy.props.IntProperty(
        name="Octaves:",
        description="",
        default = 4,
    ) 
    basis_enumerator : bpy.props.EnumProperty(
        name= "Noise Type",
        description= "Type of noise the offset has.",
        items= [
                ('PERLIN_ORIGINAL', "PERLIN_ORIGINAL", ""),
                ('BLENDER', "BLENDER", ""),
                ('PERLIN_NEW', "PERLIN_NEW", ""),
                ('VORONOI_F1', "VORONOI_F1", ""),
                ('VORONOI_F2', "VORONOI_F2", ""),
                ('VORONOI_CRACKLE', "VORONOI_CRACKLE Sphere", ""),
                ('CELLNOISE', "CELLNOISE", ""),
        ]
    )
    def draw(self, context):
        layout = self.layout
        row = layout.row()
        col = row.column()
        col2 = row.column()
        col.prop(self, "save_main_settings")
        col2.prop(self, "save_noise_settings")
        col.label(text="Main Settings")
        col.prop(self, "offset_fac")
        col.prop(self, "rot_fac")
        col.prop(self, "rot_fac_2")
        col.prop(self, "rot_vec")
        col2.label(text="Advanced Settings")
        col2.label(text="(Very Destructive)")
        col2.prop(self, "noise_offset")
        col2.prop(self,"noise_scale")
        col2.prop(self, "noise_vec")
        col2.prop(self, "basis_enumerator")
    bm = None
    @classmethod
    def bmesh(cls, context):
        me = context.edit_object.data
        cls.bm = bmesh.from_edit_mesh(me) 
        cls.bm.faces.active = None
        cls.bm.verts.ensure_lookup_table()
                
    def modal(self, context, event):
        if event.type == 'MOUSEMOVE':
            delta = self.first_mouse_y - event.mouse_y           
            self.delta = delta
            self.rot_fac = delta*0.5
            if self.bm is None or not self.bm.is_valid:
                self.bmesh(context)  
            self.execute(context)
        elif event.type == 'LEFTMOUSE':
            return {'FINISHED'}
        elif event.type in {'RIGHTMOUSE', 'ESC'}:
            if self.bm is None or not self.bm.is_valid:
                self.bmesh(context) 
            self.delta = 0 
            self.rot_fac_2 = 0
            self.rot_fac = 0
            self.rot_vec = mu.Vector((0,0,0))
            self.offset_fac = 0
            self.test_fac = 0
            self.execute(context)
            return {'CANCELLED'}
        return {'RUNNING_MODAL'}
    def invoke(self, context, event):
        if context.object:
            r3d = bpy.context.area.spaces.active.region_3d
            self.viewport = r3d.view_rotation.to_euler("XYZ")
            self.viewport_vec = mu.Vector((0.0, 0.0, 1.0))
            self.viewport_vec.rotate(self.viewport)

            if self.save_main_settings == False:
                self.rot_fac_2 = 0
                self.rot_fac = 0
                self.rot_vec = mu.Vector((0,0,0))
                self.offset_fac = 0
                self.test_fac = 0
            if self.save_noise_settings == False:
                self.noise_vec = (0,0,0)
                self.noise_scale = 1
                self.noise_offset = 0
                self.basis_enumerator = "PERLIN_ORIGINAL"

            self.bmesh(context)
            self.first_mouse_y = event.mouse_y

            mode = bpy.context.active_object.mode
            bpy.ops.object.mode_set(mode='OBJECT')
            ob = bpy.context.object 
            count = len(ob.data.vertices)
            sel = np.zeros(count, dtype=bool)
            ob.data.vertices.foreach_get('select', sel)
            sel_ind = np.arange(count)
            selected_verts = sel_ind[sel]
            bpy.ops.object.mode_set(mode=mode)
            obj = bpy.context.edit_object
            me = obj.data
            bm = bmesh.from_edit_mesh(me)
            bm.faces.active = None
            bm.verts.ensure_lookup_table()
            if len(selected_verts) < 1:
                self.report({'WARNING'}, "No proper loop found, could not finish")
                return {'CANCELLED'}  
            normala = mu.Vector((0,0,0))    
            for i in range(len(selected_verts)-1):             
                s1 = bm.verts[selected_verts[(i)%len(selected_verts)]].co
                s2 = bm.verts[selected_verts[(i+1)%len(selected_verts)]].co
                s3 = bm.verts[selected_verts[(i+2)%len(selected_verts)]].co
                delni_vektor = (s2-s1).cross((s3-s1)).normalized()  
                normala += delni_vektor
            normala = normala/len(selected_verts)
            avg_vec = mu.Vector((0,0,0))
            for index in selected_verts:
                avg_vec += bm.verts[index].co            
            avg_vec = avg_vec/len(selected_verts)
            self.me = me
            self.selected_verts = selected_verts
            self.avg_vec = avg_vec
            self.normala = normala
            self.rot_fac_2 = 0
            context.window_manager.modal_handler_add(self)
            return {'RUNNING_MODAL'}
        else:
            self.report({'WARNING'}, "No active object, could not finish")
            return {'CANCELLED'}

        return {'CANCELLED'}
    def execute(self, context):
        if self.bm is None or not self.bm.is_valid:
            self.bmesh(context)  
        rotation_angle = self.rot_fac
        rotation_angle_2 = self.rot_fac_2
        pi = 3.145926
        avg_up_vec = mu.Vector((0,0,0))
        for index in self.selected_verts:
            holypoints = []
            v = self.bm.verts[index]
            for edge_link in v.link_edges:
                p1 = edge_link.verts[0]
                p2 = edge_link.verts[1]
                if p1.index == v.index:
                    if p2.index not in self.selected_verts:
                        holypoints.append(p2.index)
                if p2.index == v.index:
                    if p1.index not in self.selected_verts:
                        holypoints.append(p1.index) 
            if len(holypoints) != 2 and len(holypoints) != 1:
                self.report({'WARNING'}, "No proper edge loop found")
                return {'CANCELLED'}                            
            s1 = self.bm.verts[self.selected_verts[0]].co
            s2 = self.bm.verts[self.selected_verts[1]].co
            s3 = self.bm.verts[self.selected_verts[2]].co
            normala_c = (s2-s1).cross((s3-s1)).normalized()
            if len(holypoints) == 1:
                r1 = self.bm.verts[holypoints[0]].co
            else:
                r1 = self.bm.verts[holypoints[0]].co
                r2 = self.bm.verts[holypoints[1]].co                 
            vec_1 = r1-v.co
            vec_1 = vec_1.project(normala_c).normalized()
            if len(holypoints) == 2: 
                vec_2 = r2-v.co
                vec_2 = vec_2.project(normala_c).normalized()
                if vec_1.dot(normala_c) > 0:
                    up_vec = r1-v.co
                else:
                    up_vec = r2-v.co
            else:
                up_vec = r1-v.co
            avg_up_vec += up_vec 
        s_normal = avg_up_vec/len(self.selected_verts)
        #mat_rot = mu.Matrix.Rotation(math.radians(self.rot_fac), 3, s_normal.orthogonal())
        mat_rot_90 = mu.Matrix.Rotation(math.radians(90), 3, s_normal)
        mat_rot = mu.Matrix.Rotation(math.radians(self.rot_fac), 3, self.viewport_vec)
        
        plane_normal = (self.normala @ mat_rot)
        mat_rot_2 = mu.Matrix.Rotation(math.radians(self.rot_fac_2), 3, s_normal) 
        plane_normal = plane_normal @ mat_rot_2
        rot_vector = mu.Vector(self.rot_vec)
        rot_vector.x = math.radians(rot_vector.x)
        rot_vector.y = math.radians(rot_vector.y)
        rot_vector.z = math.radians(rot_vector.z)
        eul = mu.Euler(rot_vector, 'XYZ')
        plane_normal.rotate(eul)
        for index in self.selected_verts:
            holypoints = []
            v = self.bm.verts[index]
            for edge_link in v.link_edges:
                p1 = edge_link.verts[0]
                p2 = edge_link.verts[1]
                if p1.index == v.index:
                    if p2.index not in self.selected_verts:
                        holypoints.append(p2.index)
                if p2.index == v.index:
                    if p1.index not in self.selected_verts:
                        holypoints.append(p1.index) 

            if len(holypoints) != 2 and len(holypoints) != 1:
                self.report({'WARNING'}, "No proper edge loop found")
                return {'CANCELLED'}                            
            s1 = self.bm.verts[self.selected_verts[0]].co
            s2 = self.bm.verts[self.selected_verts[1]].co
            s3 = self.bm.verts[self.selected_verts[2]].co
            normala_c = (s2-s1).cross((s3-s1)).normalized()
            if len(holypoints) == 1:
                r1 = self.bm.verts[holypoints[0]].co
            else:
                r1 = self.bm.verts[holypoints[0]].co   
                r2 = self.bm.verts[holypoints[1]].co                
            vec_1 = r1-v.co
            vec_1 = vec_1.project(normala_c).normalized()
            if len(holypoints) == 2: 
                vec_2 = r2-v.co
                vec_2 = vec_2.project(normala_c).normalized()
                if vec_1.dot(normala_c) > 0:
                    up_vec = r1-v.co
                else:
                    up_vec = r2-v.co
            else:
                r2 = v.co
                up_vec = r1-v.co
            p_int = mu.geometry.intersect_line_plane(r1, r2, self.avg_vec, plane_normal)
            v.co = p_int+up_vec.normalized()*self.offset_fac-up_vec*self.test_fac+up_vec.normalized()*self.test_fac
            noise_position = mu.Vector(self.noise_vec) + v.co*self.noise_scale
            v.co = v.co + up_vec*self.noise_offset*mu.noise.fractal(noise_position,self.H,self.lacunarity,self.octaves,noise_basis=self.basis_enumerator)
        bmesh.update_edit_mesh(self.me, loop_triangles=True)
        return {'FINISHED'}

def menu_func(self, context):
    self.layout.operator(ModifyGroup.bl_idname, icon='MESH_CUBE')

def register():
    bpy.utils.register_class(ModifyGroup)
    bpy.types.VIEW3D_MT_edit_mesh.append(menu_func)


def unregister():
    bpy.utils.unregister_class(ModifyGroup)
    bpy.types.VIEW3D_MT_edit_mesh.remove(menu_func)


if __name__ == "__main__":
    register()


